from .middleware import get_default_middleware
from .server_manager import ServerManager
from ..system import SystemManager

__all__ = ['serve_system']


def serve_system(middleware=None):
    system_manager = SystemManager()

    if middleware is None:
        middleware = get_default_middleware()
    server_manager = ServerManager(middleware=middleware)
    server_manager.register_endpoints(system_manager.get_default_endpoints())

    return server_manager.get_server()



from frater.factory.factory import Factory


class ObjectFactory(Factory):
    def __init__(self, base_type: type = object):
        super(ObjectFactory, self).__init__()
        self._base_type = base_type

    def get_registered_objects(self):
        return self.get_registered_values()

    def register(self, key: str):
        def wrapper(derived: type):
            if not issubclass(derived, self._base_type):
                raise TypeError(f'{derived} is not a subclass of {self._base_type}')
            elif key in self.factory_map:
                if self.factory_map[key] is not derived:
                    raise KeyError(f'{key} already exists')
            else:
                self.factory_map[key] = derived
            return derived

        return wrapper

    def register_class(self, key: str, derived: type):
        self.register(key)(derived)



from dataclasses import dataclass

from .component import Component, ComponentConfig
from ...stream import InputStream, OutputStream, is_end_of_stream, is_start_of_stream

__all__ = ['IOComponent', 'IOComponentConfig']


@dataclass
class IOComponentConfig(ComponentConfig):
    name: str = 'io_component_config'


class IOComponent(Component):
    def __init__(self, config: IOComponentConfig, input_stream: InputStream, output_stream: OutputStream):
        super(IOComponent, self).__init__(config)
        self.input_stream = input_stream
        self.output_stream = output_stream

    def run(self):
        self.start()
        for data in self.input_stream:
            if self.stopped:
                break
            if self.paused:
                self.wait()

            if is_start_of_stream(data):
                self.reset()
                self.set_active()
                self.on_start_of_stream(data)
                self.output_stream(data)
            elif is_end_of_stream(data):
                self.on_end_of_stream(data)
                self.output_stream(data)
                self.set_inactive()
            else:
                self.component_lifecycle(data)

        self.stop()

    def component_lifecycle(self, data):
        self.before_preprocess(data)
        preprocessed_input = self.preprocess(data)
        self.before_process(preprocessed_input)
        output = self.process(preprocessed_input)
        self.after_process(output)
        postprocessed_output = self.postprocess(output)
        self.after_postprocess(postprocessed_output)
        self.send_output(postprocessed_output)

    def before_preprocess(self, data):
        pass

    # noinspection PyMethodMayBeStatic
    def preprocess(self, data):
        return data

    def before_process(self, data):
        pass

    def process(self, data):
        raise NotImplementedError

    def after_process(self, data):
        pass

    # noinspection PyMethodMayBeStatic
    def postprocess(self, data):
        return data

    def after_postprocess(self, data):
        pass

    def send_output(self, output):
        self.output_stream(output)

    def on_end_of_stream(self, data):
        pass

    def on_start_of_stream(self, data):
        pass



from typing import Dict

from ...client import ComponentClientConfig, ComponentClient
from ...utilities import Handler

__all__ = ['SystemManager']


class SystemManager:
    def __init__(self):
        self.components: Dict[str, ComponentClient] = dict()

    def check_dependency(self, params):
        # TODO implement dependency checking. For now, dependencies are always ready
        return {'ready': True}

    def get_component(self, params):
        return self.components[params['component_id']].config.to_dict()

    def get_components(self):
        return [component.config.to_dict() for component in self.components.values()]

    def register_component(self, data):
        client_config = ComponentClientConfig.from_dict(data)
        if client_config.component_id in self.components:
            return {'registered': False, 'message': f'Component {client_config.component_id} already registered'}
        else:
            self.components[client_config.component_id] = ComponentClient(client_config)

            return {'registered': True, 'message': f'Component {client_config.component_id} successfully registered'}

    def start(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'started': False, 'message': f'Component {component_id} is not registered in this system.'}
        client = self.components[component_id]

        response = client.start()
        response.update({'component_id': component_id})

        return response

    def start_all(self):
        return [self.start({'component_id': component_id}) for component_id in self.components]

    def started_component(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'started': False, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.started()
        response.update({'component_id': component_id})

        return response

    def all_started(self):
        return [self.started_component({'component_id': component_id}) for component_id in self.components]

    def stop(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'stopped': False, 'message': f'Component {component_id} is not registered in this system.'}
        client = self.components[component_id]

        response = client.stop()
        response.update({'component_id': component_id})

        return response

    def stop_all(self):
        return [self.stop({'component_id': component_id}) for component_id in self.components]

    def stopped(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'stopped': False, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.stopped()
        response.update({'component_id': component_id})

        return response

    def all_stopped(self):
        return [self.stopped({'component_id': component_id}) for component_id in self.components]

    def toggle_pause(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'paused': None, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.toggle_pause()
        response.update({'component_id': component_id})

        return response

    def toggle_pause_all(self):
        return [self.toggle_pause({'component_id': component_id}) for component_id in self.components]

    def pause(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'paused': False, 'message': f'Component {component_id} is not registered in this system.'}
        client = self.components[component_id]

        response = client.pause()
        response.update({'component_id': component_id})

        return response

    def pause_all(self):
        return [self.pause({'component_id': component_id}) for component_id in self.components]

    def unpause(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'unpaused': False, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.stopped()
        response.update({'component_id': component_id})

        return response

    def unpause_all(self):
        return [self.unpause({'component_id': component_id}) for component_id in self.components]

    def paused(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'paused': None, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.paused()
        response.update({'component_id': component_id})

        return response

    def all_paused(self):
        return [self.paused({'component_id': component_id}) for component_id in self.components]

    def active(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'active': None, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.active()
        response.update({'component_id': component_id})

        return response

    def all_active(self):
        return [self.active({'component_id': component_id}) for component_id in self.components]

    def reset(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'reset': False, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.reset()
        response.update({'component_id': component_id})

        return response

    def reset_all(self):
        return [self.reset({'component_id': component_id}) for component_id in self.components]

    def error(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'error': False, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.error()
        response.update({'component_id': component_id})

        return response

    def all_error(self):
        return [self.error({'component_id': component_id}) for component_id in self.components]

    def update_config(self, data):
        component_id = data['component_id']
        config = data['config']
        if component_id not in self.components:
            return {'updated': False, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.update_config(config)
        response.update({'component_id': component_id})

        return response

    def get_config(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'stopped': False, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.get_config()

        return response

    def get_all_configs(self):
        return [self.get_config({'component_id': component_id}) for component_id in self.components]

    def get_state(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'stopped': False, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = {'component_id': component_id, 'state': client.get_status()}

        return response

    def get_all_states(self):
        return [self.get_state({'component_id': component_id}) for component_id in self.components]

    def get_status(self, params):
        component_id = params['component_id']
        if component_id not in self.components:
            return {'stopped': False, 'message': f'Component {component_id} is not registered in this system.'}

        client = self.components[component_id]

        response = client.get_status()
        response.update({'component_id': component_id})

        return response

    def get_all_statuses(self):
        return [self.get_status({'component_id': component_id}) for component_id in self.components]

    def get_default_endpoints(self):
        return [
            Handler('/dependency', 'check_dependency', self.check_dependency),
            Handler('/component', 'get_component', self.get_component),
            Handler('/component/register', 'register_component', self.register_component, methods=['POST']),
            Handler('/component/start', 'start', self.start),
            Handler('/component/started', 'started_component', self.started_component),
            Handler('/component/stop', 'stop', self.stop),
            Handler('/component/stopped', 'stopped', self.stopped),
            Handler('/component/toggle_pause', 'toggle_pause', self.toggle_pause),
            Handler('/component/pause', 'pause', self.pause),
            Handler('/component/unpause', 'unpause', self.unpause),
            Handler('/component/paused', 'paused', self.paused),
            Handler('/component/active', 'active', self.active),
            Handler('/component/reset', 'reset', self.reset),
            Handler('/component/error', 'error', self.error),
            Handler('/component/config', 'get_config', self.get_config),
            Handler('/component/config', 'update_config', self.update_config, methods=['PATCH']),
            Handler('/component/state', 'get_state', self.get_state),
            Handler('/component/status', 'get_status', self.get_status),
            Handler('/components', 'get_components', self.get_components),
            Handler('/components/start', 'start_all', self.start_all),
            Handler('/components/started', 'all_started', self.all_started),
            Handler('/components/stop', 'stop_all', self.stop_all),
            Handler('/components/stopped', 'all_stopped', self.all_stopped),
            Handler('/components/toggle_pause', 'toggle_pause_all', self.toggle_pause_all),
            Handler('/components/pause', 'pause_all', self.pause_all),
            Handler('/components/unpause', 'unpause_all', self.unpause_all),
            Handler('/components/paused', 'all_paused', self.all_paused),
            Handler('/components/active', 'all_active', self.all_active),
            Handler('/components/reset', 'reset_all', self.reset_all),
            Handler('/components/error', 'all_error', self.all_error),
            Handler('/components/config', 'get_all_configs', self.get_all_configs),
            Handler('/components/state', 'get_all_states', self.get_all_states),
            Handler('/components/status', 'get_all_statuses', self.get_all_statuses),
        ]



import json


def is_json_serializable(x):
    try:
        json.dumps(x)
        return True
    except:
        return False


def is_json_deserializable(x, decode_type='utf-8'):
    try:
        json.loads(x.decode())
        return True
    except:
        return False



import json
from typing import Callable

from ..data_type import DataType
from ..factory import Factory
from ..utilities.json import is_json_serializable

__all__ = ['json_serializers', 'frater_to_json', 'get_kafka_serializer']

json_serializers = Factory()


def frater_to_json(data):
    d_type = type(data)

    if isinstance(data, list):
        return [frater_to_json(item) for item in data]
    elif issubclass(d_type, DataType):
        return data.to_dict()
    elif d_type in json_serializers:
        return json_serializers[d_type](data)
    elif is_json_serializable(data):
        return data
    else:
        raise TypeError(f'Object can\'t be serialized to json: {data}')


def get_kafka_serializer() -> Callable:
    return lambda m: json.dumps(frater_to_json(m)).encode('utf-8')



from dataclasses import dataclass, field
from typing import List
from uuid import uuid4

from dataclasses_json import DataClassJsonMixin

from ...config import Config
from ...dependency import Dependency
from ...utilities import Handler

__all__ = ['Component', 'ComponentConfig', 'ComponentState']


@dataclass
class ComponentConfig(Config):
    """Base
    :param component_id: Base :py:class:`~frater.config.Config` class for components
    """
    component_id: str = field(default_factory=lambda: str(uuid4()))
    name: str = 'component_config'
    dependencies: List[Dependency] = field(default_factory=list)


@dataclass
class ComponentState(DataClassJsonMixin):
    """
    :param state_id: test
    :type state_id: str
    """
    state_id: str = field(default_factory=lambda: str(uuid4()))


class Component:
    def __init__(self, config: ComponentConfig = None):
        """Base class for components in Frater. Generally shouldn't be used for direct subclassing, outside of\
        advanced use cases. In general, use :py:class:`~frater.component.component.io_component.IOComponent`,
        :py:class:`~frater.component.component.output_component.OutputComponent`, \
        :py:class:`~frater.component.component.input_component.InputComponent` or their subclasses as starting points.

        :param config: config object for the component. Should be a :py:class:`~frater.component.ComponentConfig` or a \
        subclass with more fields.
        """
        if config is None:
            config = ComponentConfig()

        self.started = False
        self.paused = False
        self.active = False

        self.config = config
        self.state = self.init_state()

    @property
    def stopped(self):
        return not self.started

    def init_state(self) -> ComponentState:
        return ComponentState()

    def run(self):
        raise NotImplementedError

    def reset(self):
        self.state = self.init_state()

    def start(self):
        self.started = True

    def stop(self):
        self.set_inactive()
        self.started = False

    def pause(self):
        self.paused = True

    def unpause(self):
        self.paused = False

    def toggle_pause(self):
        self.paused = not self.paused

    def set_active(self):
        self.active = True

    def set_inactive(self):
        self.active = False

    def wait(self):
        while self.paused:
            continue

    # noinspection PyMethodMayBeStatic
    def get_additional_handlers(self) -> List[Handler]:
        return []



from threading import Thread
from typing import Optional

from ..component import Component


class ComponentThread(Thread):
    def __init__(self, component: Component):
        super(ComponentThread, self).__init__()
        self.component = component

        self.error = None
        self.started_once = False

    def run(self):
        try:
            self.component.run()
        except BaseException as e:
            self.error = e

        self.component.stop()

    def join(self, timeout: Optional[float] = ...) -> None:
        super(ComponentThread, self).join()
        if self.has_error():
            raise self.error

    def has_error(self):
        return self.error is not None

    def start(self):
        super(ComponentThread, self).start()
        self.started_once = True



from dataclasses import dataclass
from typing import Dict

import inflection
from dataclasses_json import DataClassJsonMixin
from dataclasses_json.core import Json


@dataclass
class DataType(DataClassJsonMixin):
    @classmethod
    def data_type(cls):
        return inflection.underscore(cls.__name__)

    def to_dict(self, encode_json=False) -> Dict[str, Json]:
        d = super(DataType, self).to_dict()
        d['data_type'] = self.data_type()

        return d



import requests


class API:
    def __init__(self, host, port, protocol='http://'):
        self.host = host
        self.port = port
        self.protocol = protocol

    @property
    def api_url(self):
        return f'{self.protocol}{self.host}:{self.port}'

    def build_endpoint_url(self, endpoint: str):
        if not endpoint.startswith('/'):
            endpoint = '/' + endpoint

        return f'{self.api_url}{endpoint}'

    def get(self, endpoint: str, params=None):
        endpoint_url = self.build_endpoint_url(endpoint)
        response = requests.get(endpoint_url, params=params)
        return response.json()

    def post(self, endpoint: str, data: dict):
        endpoint_url = self.build_endpoint_url(endpoint)
        response = requests.post(endpoint_url, json=data)
        return response.json()

    def delete(self, endpoint: str, data: dict):
        endpoint_url = self.build_endpoint_url(endpoint)
        response = requests.delete(endpoint_url, json=data)
        return response.json()

    def patch(self, endpoint: str, data: dict):
        endpoint_url = self.build_endpoint_url(endpoint)
        response = requests.patch(endpoint_url, json=data)
        return response.json()

    def put(self, endpoint: str, data: dict):
        endpoint_url = self.build_endpoint_url(endpoint)
        response = requests.put(endpoint_url, json=data)
        return response.json()



import logging
from dataclasses import field, dataclass

from .handler.factory import get_handler
from ..config import Config


@dataclass
class LoggerConfig(Config):
    name: str = field(default=__name__)
    handler: str = 'stdout'
    format: str = '%(asctime)s | %(name)s | %(levelname)s | %(message)s'
    level: int = logging.INFO


def get_logger(config: LoggerConfig) -> logging.Logger:
    handler = get_handler(config.handler)
    formatter = logging.Formatter(config.format)
    handler.setFormatter(formatter)

    logger = logging.getLogger(config.name)
    logger.addHandler(handler)
    logger.setLevel(config.level)

    return logger



from dataclasses import dataclass, field
from typing import List

from .stream import InputStream, OutputStream, StreamConfig
from ..config import Config
from ..dependency import Dependency
from ..factory import ObjectFactory

__all__ = ['InputStreamFactoryConfig', 'OutputStreamFactoryConfig',
           'StreamFactoryConfig', 'input_stream_factory', 'input_stream_configs',
           'output_stream_factory', 'output_stream_configs']

input_stream_factory = ObjectFactory(InputStream)
input_stream_configs = ObjectFactory(StreamConfig)

output_stream_factory = ObjectFactory(OutputStream)
output_stream_configs = ObjectFactory(StreamConfig)


@dataclass
class InputStreamFactoryConfig(Config):
    name: str = 'input_stream_factory_config'
    stream_type: str = ''
    data_type: str = ''
    stream_config: dict = field(default_factory=dict)
    dependencies: List[Dependency] = field(default_factory=list)


@dataclass
class OutputStreamFactoryConfig(Config):
    name: str = 'output_stream_factory_config'
    stream_type: str = ''
    data_type: str = ''
    stream_config: dict = field(default_factory=dict)
    dependencies: List[Dependency] = field(default_factory=list)


@dataclass
class StreamFactoryConfig(Config):
    name: str = 'stream_factory_config'
    input_stream_config: InputStreamFactoryConfig = field(default_factory=InputStreamFactoryConfig)
    output_stream_configs: List[OutputStreamFactoryConfig] = field(default_factory=list)

    @property
    def dependencies(self):
        dependencies = list()
        dependencies.extend(self.input_stream_config.dependencies)
        for config in self.output_stream_configs:
            dependencies.extend(config.dependencies)

        return dependencies



class Factory:
    def __getitem__(self, key):
        return self.get(key)

    def __init__(self):
        self.factory_map = dict()

    def __contains__(self, item):
        return item in self.factory_map

    def register(self, key):
        def wrapper(value):
            if key in self.factory_map:
                if self.factory_map[key] is not value:
                    raise KeyError(f'{key} already exists')
            else:
                self.factory_map[key] = value
            return value

        return wrapper

    def register_item(self, key, value):
        return self.register(key)(value)

    def unregister(self, key):
        if key not in self.factory_map:
            raise KeyError(f'{key} does not exist')

        del self.factory_map[key]

    def get(self, key):
        return self.factory_map[key]

    def get_registered_keys(self):
        return list(self.factory_map.keys())

    def get_registered_values(self):
        return list(self.factory_map.values())



from dataclasses import dataclass, field
from typing import List

from kafka import KafkaProducer, KafkaConsumer

from .factory import *
from .stream import OutputStream, InputStream, StreamConfig
from ..io import get_kafka_serializer, get_kafka_deserializer


@input_stream_configs.register('kafka')
@dataclass
class KafkaInputStreamConfig(StreamConfig):
    topics: List[str] = field(default_factory=list)
    servers: List[str] = field(default_factory=lambda: ['localhost:9092'])


@input_stream_factory.register('kafka')
class KafkaInputStream(InputStream):
    def __init__(self, config: KafkaInputStreamConfig):
        super(KafkaInputStream, self).__init__(config)

        deserializer = get_kafka_deserializer()

        self._consumer = KafkaConsumer(*self.config.topics, bootstrap_servers=self.config.servers,
                                       value_deserializer=deserializer)

    def __next__(self):
        return next(self._consumer).value

    def __iter__(self):
        for msg in self._consumer:
            yield msg.value

    def close(self):
        self._consumer.close()


@output_stream_configs.register('kafka')
@dataclass
class KafkaOutputStreamConfig(StreamConfig):
    topic: str = ''
    servers: List[str] = field(default_factory=lambda: ['localhost:9092'])


@output_stream_factory.register('kafka')
class KafkaOutputStream(OutputStream):
    def __init__(self, config: KafkaOutputStreamConfig):
        super(KafkaOutputStream, self).__init__(config)

        serializer = get_kafka_serializer()
        self._producer = KafkaProducer(bootstrap_servers=self.config.servers,
                                       value_serializer=serializer)

    def send(self, data):
        self._producer.send(self.config.topic, data).get()

    def close(self):
        self._producer.close()



from typing import Tuple

from .factory import *
from .stream import InputStream, OutputStream, MultiOutputStream


def get_streams(config: StreamFactoryConfig) -> Tuple[InputStream, OutputStream]:
    input_stream = get_input_stream(config.input_stream_config)

    if len(config.output_stream_configs) == 0:
        output_stream = None
    elif len(config.output_stream_configs) > 1:
        output_streams = [get_output_stream(stream_config) for stream_config in config.output_stream_configs]
        output_stream = MultiOutputStream(output_streams)
    else:
        output_stream = get_output_stream(config.output_stream_configs[0])

    return input_stream, output_stream


def get_input_stream(config: InputStreamFactoryConfig) -> InputStream:
    if not config:
        return None
    else:
        stream_config = get_input_stream_config(config)
        return input_stream_factory[config.stream_type](stream_config)


def get_input_stream_config(config: InputStreamFactoryConfig):
    config_class = input_stream_configs[config.stream_type]
    return config_class.from_dict(config.stream_config)


def get_output_stream(config: OutputStreamFactoryConfig) -> OutputStream:
    if not config:
        return None
    else:
        stream_config = get_output_stream_config(config)
        return output_stream_factory[config.stream_type](stream_config)


def get_output_stream_config(config: OutputStreamFactoryConfig):
    config_class = output_stream_configs[config.stream_type]
    return config_class.from_dict(config.stream_config)



from dataclasses import dataclass

from .component import Component, ComponentConfig
from ...stream import OutputStream

__all__ = ['OutputComponent', 'OutputComponentConfig']


@dataclass
class OutputComponentConfig(ComponentConfig):
    name: str = 'output_component_config'


class OutputComponent(Component):
    def __init__(self, config: ComponentConfig, output_stream: OutputStream):
        super(OutputComponent, self).__init__(config)
        self.output_stream = output_stream

    def run(self):
        self.start()
        while self.should_output() and not self.stopped:
            self.component_lifecycle()
        self.stop()

    def component_lifecycle(self):
        self.before_process()
        output = self.process()
        self.after_process(output)
        postprocessed_output = self.postprocess(output)
        self.after_postprocess(postprocessed_output)
        self.send_output(postprocessed_output)

    def before_process(self):
        pass

    def process(self):
        raise NotImplementedError

    def after_process(self, data):
        pass

    # noinspection PyMethodMayBeStatic
    def postprocess(self, data):
        return data

    def after_postprocess(self, data):
        pass

    # noinspection PyMethodMayBeStatic
    def should_output(self):
        return True

    def send_output(self, output):
        self.output_stream(output)



import os
from dataclasses import dataclass

from ...config import Config
from ...utilities import Singleton


@dataclass
class FileStoreConfig(Config):
    name: str = 'file_store_config'
    root: str = ''


class FileStore(metaclass=Singleton):
    def __init__(self, config: FileStoreConfig):
        self.config = config

    @property
    def root(self):
        return os.path.abspath(os.path.expandvars(os.path.expanduser(self.config.root)))

    def get_full_path(self, file_path):
        return os.path.join(self.root, file_path)



from threading import Lock
from typing import List

from .thread import ComponentThread
from ..component import Component
from ...utilities import Handler


class ComponentManager:
    def __init__(self, component: Component):
        self.component = component
        self.lock = Lock()
        self.thread = self.create_component_thread()

    def create_component_thread(self):
        return ComponentThread(self.component)

    def started(self):
        return {'started': self.component.started}

    def stopped(self):
        return {'stopped': self.component.stopped}

    def start(self):
        if self.component.started:
            message = 'Component already started. Cannot be started twice.'
        else:
            if self.thread.started_once:
                self.thread = self.create_component_thread()
                message = 'Component restarted.'
            else:
                message = 'Component successfully started.'

            self.thread.start()

        return {'started': True, 'message': message}

    def stop(self):
        if self.component.paused:
            return {'stopped': False, 'message': 'unable to stop when component is paused'}

        if self.component.stopped:
            return {'stopped': True, 'message': 'Component already stopped'}

        with self.lock:
            self.component.stop()

        try:
            self.thread.join()
            return {'stopped': True, 'message': 'Component successfully stopped'}

        except BaseException as e:
            return {'stopped': False, 'message': str(e)}

    def error(self):
        has_error = self.thread.has_error()
        message = str(self.thread.error) if has_error else 'No current error'
        response = {'error': has_error, 'message': message}

        return response

    def reset(self):
        with self.lock:
            self.component.reset()

        return {'reset': True}

    def get_config(self):
        return self.component.config.to_dict()

    def update_config(self, update):
        with self.lock:
            self.component.config.update_from_dict(update)

        return {'updated': True, 'config': self.component.config.to_dict()}

    def toggle_pause(self):
        if self.component.stopped:
            return {'paused': None, 'error': True, 'message': 'unable to pause since component is not started'}
        with self.lock:
            self.component.toggle_pause()

        return {'paused': self.component.paused, 'message': 'Pause toggled'}

    def pause(self):
        if self.component.stopped:
            return {'paused': False, 'message': 'unable to pause since component is not started'}
        if self.component.paused:
            return {'paused': True, 'message': 'Component already paused'}
        else:
            with self.lock:
                self.component.pause()

            return {'paused': self.component.paused}

    def unpause(self):
        if self.component.stopped:
            return {'unpaused': False, 'message': 'unable to unpause since component is not started'}
        if not self.component.paused:
            return {'unpaused': True, 'message': 'Component already unpaused'}
        else:
            with self.lock:
                self.component.unpause()
            return {'unpaused': not self.component.paused}

    def paused(self):
        return {'paused': self.component.paused}

    def active(self):
        return {'active': self.component.active}

    def get_status(self):
        status = {
            'config': self.get_config(),
            'state': self.get_state()
        }

        status.update(self.started())
        status.update(self.active())
        status.update(self.stopped())
        status.update(self.paused())

        return status

    def get_state(self):
        return self.component.state.to_dict()

    def get_endpoints(self):
        return [handler.to_dict() for handler in self.get_all_handlers()]

    def get_all_handlers(self):
        return self.get_default_handlers() + self.component.get_additional_handlers()

    def get_default_handlers(self) -> List[Handler]:
        return [
            Handler('/start', 'start', self.start),
            Handler('/started', 'started', self.started),
            Handler('/stop', 'stop', self.stop),
            Handler('/stopped', 'stopped', self.stopped),
            Handler('/toggle_pause', 'toggle_pause', self.toggle_pause),
            Handler('/pause', 'pause', self.pause),
            Handler('/unpause', 'unpause', self.unpause),
            Handler('/paused', 'paused', self.paused),
            Handler('/active', 'active', self.active),
            Handler('/reset', 'reset', self.reset),
            Handler('/error', 'error', self.error),
            Handler('/config', 'update_config', self.update_config, methods=['PATCH']),
            Handler('/config', 'get_config', self.get_config),
            Handler('/state', 'get_state', self.get_state),
            Handler('/status', 'get_status', self.get_status),
            Handler('/endpoints', 'get_endpoints', self.get_endpoints)
        ]



import time
from dataclasses import dataclass, field
from pprint import pformat
from typing import List, Dict

from .api import API
from .component import ComponentClientConfig
from ..config import Config
from ..dependency import Dependency
from ..utilities import NetworkConfig

__all__ = ['SystemClient', 'SystemClientConfig']


@dataclass
class SystemClientConfig(Config):
    name: str = 'system_client_config'
    network_config: NetworkConfig = field(default_factory=NetworkConfig)
    polling_timeout: int = 3
    max_retries: int = 10


class SystemClient:
    def __init__(self, config: SystemClientConfig):
        self.system_api = API(config.network_config.host, config.network_config.port)
        self.config = config

    def wait_for_dependencies(self, dependencies: List[Dependency]):
        if len(dependencies) == 0:
            return

        all_ready = False
        retries = 0

        while not all_ready:
            dependency_statuses = [self.check_dependency(dependency) for dependency in dependencies]
            all_ready = all(status['ready'] for status in dependency_statuses)

            if not all_ready:
                retries += 1
                if retries >= self.config.max_retries:
                    raise TimeoutError(f'Took too long waiting for dependencies: {pformat(dependency_statuses)}')

                time.sleep(self.config.polling_timeout)

    def check_dependency(self, dependency: Dependency):
        return self.system_api.get('/dependency', params=dependency.to_dict())

    def get_component(self, component_id: str):
        return self.system_api.get('/component', params={'component_id': component_id})

    def register_component(self, component_client_config: ComponentClientConfig):
        response = self.system_api.post('/component/register', component_client_config.to_dict())
        return response

    def start(self, component_id: str):
        return self.system_api.get('/component/start', params={'component_id': component_id})

    def started(self, component_id: str):
        return self.system_api.get('/component/started', params={'component_id': component_id})

    def stop(self, component_id: str):
        return self.system_api.get('/component/stop', params={'component_id': component_id})

    def stopped(self, component_id: str):
        return self.system_api.get('/component/stopped', params={'component_id': component_id})

    def toggle_pause(self, component_id: str):
        return self.system_api.get('/component/toggle_pause', params={'component_id': component_id})

    def pause(self, component_id: str):
        return self.system_api.get('/component/pause', params={'component_id': component_id})

    def unpause(self, component_id: str):
        return self.system_api.get('/component/unpause', params={'component_id': component_id})

    def paused(self, component_id: str):
        return self.system_api.get('/component/paused', params={'component_id': component_id})

    def active(self, component_id: str):
        return self.system_api.get('/component/active', params={'component_id': component_id})

    def reset(self, component_id: str):
        return self.system_api.get('/component/reset', params={'component_id': component_id})

    def error(self, component_id: str):
        return self.system_api.get('/component/error', params={'component_id': component_id})

    def get_config(self, component_id: str):
        return self.system_api.get('/component/config', params={'component_id': component_id})

    def update_config(self, component_id: str, config: Dict):
        return self.system_api.patch('/component/config', data={'component_id': component_id, 'config': config})

    def get_state(self, component_id: str):
        return self.system_api.get('/component/state', params={'component_id': component_id})

    def get_status(self, component_id: str):
        return self.system_api.get('/component/status', params={'component_id': component_id})

    def get_components(self):
        return self.system_api.get('/components')

    def start_all(self):
        return self.system_api.get('/components/start')

    def all_started(self):
        return self.system_api.get('/components/started')

    def stop_all(self):
        return self.system_api.get('/components/stop')

    def all_stopped(self):
        return self.system_api.get('/components/stopped')

    def toggle_pause_all(self):
        return self.system_api.get('/components/toggle_pause')

    def pause_all(self):
        return self.system_api.get('/components/pause')

    def unpause_all(self):
        return self.system_api.get('/components/unpause')

    def all_paused(self):
        return self.system_api.get('/components/paused')

    def all_active(self):
        return self.system_api.get('/components/active')

    def reset_all(self):
        return self.system_api.get('/components/reset')

    def all_error(self):
        return self.system_api.get('/components/error')

    def get_all_configs(self):
        return self.system_api.get('/components/config')

    def get_all_states(self):
        return self.system_api.get('/components/state')

    def get_all_statuses(self):
        return self.system_api.get('/components/status')



import os
from dataclasses import dataclass

from PIL import Image

from .file_store import FileStore, FileStoreConfig
from ...utilities.image import load_image_from_url, load_image_from_file, save_image_to_file
from ...utilities.url import is_url


@dataclass
class ImageStoreConfig(FileStoreConfig):
    name: str = 'image_store_config'


class ImageStore(FileStore):
    def __init__(self, root):
        super(ImageStore, self).__init__(root)

    def load_image(self, image_path: str) -> Image.Image:
        if is_url(image_path):
            return load_image_from_url(image_path)

        if os.path.isfile(image_path):
            return load_image_from_file(image_path)

        full_path = self.get_full_path(image_path)
        if os.path.isfile(full_path):
            return load_image_from_file(full_path)
        else:
            raise FileNotFoundError(f'Unable to find image at {full_path}')

    def save_image(self, image: Image.Image, image_path: str):
        if os.path.isdir(os.path.dirname(image_path)):
            save_image_to_file(image, image_path)
        else:
            full_path = self.get_full_path(image_path)
            save_image_to_file(image, full_path)



import logging
from typing import Dict

from .config import HandlerConfig
from ...factory import ObjectFactory

__all__ = ['get_handler', 'get_handler_config', 'handler_factory', 'handler_config_factory']

handler_factory = ObjectFactory(logging.Handler)
handler_config_factory = ObjectFactory(HandlerConfig)


def get_handler(handler: str, handler_config: Dict = None) -> logging.Handler:
    config = get_handler_config(handler, handler_config)
    return handler_factory.get(handler)(config)


def get_handler_config(handler: str, handler_config: Dict = None):
    if handler_config is None:
        handler_config = dict()
    config = handler_config_factory.get(handler).from_dict(handler_config)
    return config
